package gov.lbl.kdd.data;

import gov.ornl.csed.csiir.idr.kdd.dataservice.client.DataConsumer;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.DataSetQueryRequest;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.ResultSetResponse;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.ResultSetResponse.ResultRecord;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.ResultSetResponse.ResultRecord.ResultField;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

public class DataConsumerImpl implements DataConsumer {
	private static Logger logger = Logger.getLogger(DataConsumerImpl.class);
	
	private int run;
	private int total;
	private int count;
	private int errors;
	private int inconsistent;
	private int exceptions;
	private Object listener;
	private Map<String, List<ResultSetResponse>> dataMap;
	private Map<String, List<ResultSetResponse>> firstDataMap = Collections.synchronizedMap(new TreeMap<String, List<ResultSetResponse>>());
	private Map<String, String> dataRequestMap = new HashMap<String, String>();
	private Map<String, Long> processingTimeMap;
	
	private final String name;
	
	
	public DataConsumerImpl(String name, List<DataSetQueryRequest> dataRequests, Object listener) {
		this.name = name;
		this.total = dataRequests.size();
		this.listener = listener;
		
		if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
			for (DataSetQueryRequest dataRequest: dataRequests) {
				dataRequestMap.put(dataRequest.getQueryId(), dataRequest.getQueue().toString());
			}
		}
	}
	
	public void data(String arg0, String arg1, ResultSetResponse arg2) {
		String key = arg0 + ":" + arg1;
		
		logger.info("PERF:RUN=" + run + ":KEY="+ key + ":NUM_OF_RECORDS=" + arg2.recordSize() + ":PROCESSED_IN=" + arg2.getQueryProcessingTime());
		
		if (processingTimeMap.get(key) != null) {
			processingTimeMap.put(key, Long.valueOf(processingTimeMap.get(key) + arg2.getQueryProcessingTime()));
		} else {
			processingTimeMap.put(key, Long.valueOf(arg2.getQueryProcessingTime()));
		}
	
		if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
			if (dataMap.get(key) != null) {
				List<ResultSetResponse> list = dataMap.get(key);
				
				list.add(arg2);
			} else {
				List<ResultSetResponse> list = new ArrayList<ResultSetResponse>();
				
				list.add(arg2);
				dataMap.put(key, list);
			} 
		}
	}
	
	public void done(String arg0, String arg1) {
		String key = arg0 + ":" + arg1;
		
		if (processingTimeMap.get(key) != null) {
			logger.info("PERF:RUN="+ run + ":KEY="+ key + ":PROCESSING_TIME=" + processingTimeMap.get(key));
		} else {
			logger.info("PERF:RUN="+ run + ":KEY="+ key + " No processing time ...");
		}
		
		try {
			if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
				if (run == 1) {
					firstDataMap.put(key, dataMap.get(key));
				} else {
					List<ResultSetResponse> firstList = firstDataMap.get(key);
					List<ResultSetResponse> newList = dataMap.get(key);
					
					if (firstList == null && newList != null) {
						logger.error("ERROR_INCONSISTENT:RUN=" + run + ":KEY=" + key + ":" + firstList + ":" + newList.size());
						inconsistent++;
					} if (firstList != null && newList == null) {
						logger.error("ERROR_INCONSISTENT:RUN" + run + ":KEY=" + key + ":" + firstList.size() + ":" + newList);
						inconsistent++;
					} else if (firstList != null && firstList.size() != newList.size()) {
						logger.error("ERROR_INCONSISTENT:RUN" + run + ":KEY=" + key + ":" + firstList.size() + ":" + newList.size());
						inconsistent++;
					} else if (firstList != null) {
						for (int i = 0; i < firstList.size(); ++i) {
							if (firstList.get(i).getColumnList().equals(newList.get(i).getColumnList()) == false) {
								logger.error("ERROR_INCONSISTENT:INTERNAL_COLUMNS:RUN=" + run + ":KEY=" + key);
								inconsistent++;
								break;
							}
						}
					}
				}
			}
		} catch(Exception e) {
			logger.error("PERF:SEVERE", e);
		}
		
		synchronized(this) {
			count++;
			
			if (count == total) {
				if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
					PrintWriter writer = null;
					
					try {
						writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream("results.log") ,"utf-8"));

						for (Map.Entry<String, List<ResultSetResponse>> e: firstDataMap.entrySet()) {
							if (e.getKey().indexOf(arg0 + ":") == -1) {
								throw new RuntimeException("should never happen");
							}
							
							String queryId = e.getKey().substring(e.getKey().indexOf(':') + 1);
							writer.print("DATASET_ID=");
							writer.print(arg0);
							writer.print(":REQUEST_ID=");
							writer.print(queryId);
							writer.print(":QUEUE=");
							writer.println(dataRequestMap.get(queryId));
						
							
							if (e.getValue() != null) {
								for (ResultSetResponse res: e.getValue()) {
									
									for (ResultRecord rc: res) {
										writer.println("\tRecord:");
										
										for (ResultField rf: rc) {
											writer.print("\t\t");
											writer.print(rf.getFieldName());
											writer.print(":");
										    writer.println(rf.getObject() != null? rf.getObject().toString().trim() : null);
										}
									}
								} 
							} else {
								writer.println("\tNo Records were returned.");
							}
						} 
						
						writer.flush();
					} catch(IOException ioe) {
						logger.error("PERF:IO:OCCURED WHEN DUMPING RESULTS");
					} catch(Exception e) {
						logger.error("PERF:SEVERE:OCCURED WHEN DUMPING RESULTS", e);
					} finally {
						if (writer != null) {
							writer.close();
							
							if (writer.checkError()) {
								logger.error("PERF:IOE:OCCURED WHEN DUMPING RESULTS");
							}
						}
					}
				}
				
				synchronized(listener) {
					listener.notify();
				}
			}
		}
	}
	
	public void exception(String arg0, String arg1, Exception arg2) {
		if (logger.isDebugEnabled()) {
			logger.debug("ERROR_CALLBACK_EXCEPTION:" + arg0 + ":" + arg1 + arg2.getMessage(), arg2);
		} else {
			logger.error("ERROR_CALLBACK_EXCEPTION:" + arg0 + ":" + arg1 + arg2.getMessage());
		}
		
		synchronized(this) {
			exceptions++;
		}
	}

	public void httpResponseCode(String arg0, String arg1, int arg2) {
		if (arg2 != 200) {
			logger.error("ERROR_HTTP_CODE:" + arg0 + ":" + arg1 + ":HTTP_CODE=" + arg2);
			
			synchronized(this) {
				errors++;
			}
		}
	}

	public boolean isDone() {
		return count == total;
	}
	
	public String toString() {
		return "NAME=" + name + " [COUNT=" + count + ":TOTAL=" + total 
				+ ":HTTP_CODE_NOT_OK=" + errors + ":INCONSISTENT=" + this.inconsistent + ":CALLBACK_EXCEPTIONS=" + exceptions + "]";
	}
	
	public void init() {
		count = 0;
		errors = 0;
		inconsistent = 0;
		exceptions = 0;
		dataMap = Collections.synchronizedMap(new HashMap<String, List<ResultSetResponse>>());
		processingTimeMap = Collections.synchronizedMap(new HashMap<String, Long>());
		run++;
	}

	public int getRun() {
		return run;
	}
	
	public int getCount() {
		return count;
	}

	public int getErrors() {
		return errors;
	}

	public int getInconsistent() {
		return inconsistent;
	}

	public int getExceptions() {
		return exceptions;
	}
}
