package gov.lbl.kdd.data;

import gov.ornl.csed.csiir.idr.kdd.dataservice.client.RecordCountConsumer;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.DataSetQueryRequest;
import gov.ornl.csed.csiir.idr.kdd.dataservice.server.models.RecordCountResponse;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;

public class RecordCountConsumerImpl implements RecordCountConsumer  {
	private static Logger logger = Logger.getLogger(RecordCountConsumerImpl.class);
	
	private int run;
	private int total;
	private int count;
	private int errors;
	private int inconsistent;
	private int exceptions;
	private Object listener;
	private final Map<String, Long> firstTotalMatchingRecordMap = Collections.synchronizedMap(new TreeMap<String, Long>());
	private Map<String, Long> totalMatchingRecordMap;
	private final Map<String, String> dataRequestMap = new HashMap<String, String>();
	private final String name;

	public RecordCountConsumerImpl(String name, List<DataSetQueryRequest> dataRequests, Object listener) {
		this.name = name;
		this.total = dataRequests.size();
		this.listener = listener;
		
		if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
			for (DataSetQueryRequest dataRequest: dataRequests) {
				dataRequestMap.put(dataRequest.getQueryId(), dataRequest.getQueue().toString());
			}
		}
	}
	
	public void recordCount(String arg0, String arg1, RecordCountResponse arg2) {
		String key = arg0 + ":" + arg1;
		
		logger.info("PERF:RUN=" + run + ":KEY=" + key + ":MATCHING_RECORDS=" + arg2.getTotalMatchingRecords() 
				+ ":PROCESSING_TIME=" + arg2.getQueryProcessingTime());
		
		if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
			if (totalMatchingRecordMap.get(key) == null) {
				totalMatchingRecordMap.put(key, arg2.getTotalMatchingRecords());
		    } else {
		    	totalMatchingRecordMap.put(key, Long.valueOf(totalMatchingRecordMap.get(key) + arg2.getTotalMatchingRecords()));
		    }
		}
	}
	
	public void done(String arg0, String arg1) {
		String key = arg0 + ":" + arg1;
		
		logger.info("PERF:RUN=" + run + ":KEY=" + key +
				(totalMatchingRecordMap.get(key) == null? ":No records" : "TOTAL_MATCHING_RECORDS=" + totalMatchingRecordMap.get(key)));
		
		if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
			if (run == 1) {
				firstTotalMatchingRecordMap.put(key, totalMatchingRecordMap.get(key));
			} else {
				Long firstSize = firstTotalMatchingRecordMap.get(key);
				Long newSize = totalMatchingRecordMap.get(key);
				
				if (logger.isDebugEnabled()) {
					logger.debug(key + ":" + firstSize + ":" + newSize);
				}
				
				if (firstSize == null && newSize != null) {
					logger.error("ERROR_INCONSISTENT:RUN=" + run + ":KEY=" + key + ":" + firstSize + ":" + newSize);
					inconsistent++;
				} else if (firstSize != null && newSize == null) {
					logger.error("ERROR_INCONSISTENT:RUN=" + key + ":" + firstSize + ":" + newSize);
					inconsistent++;
				} else if (firstSize != null && firstSize.equals(newSize) == false) {
					logger.error("ERROR_INCONSISTENT:RUN=" + run + ":KEY=" + key + ":" + firstSize + ":" + newSize);
					inconsistent++;
				} 
			}
		}
		
		synchronized(this) {
			count++;
			
			if (count == total) {
				if ("true".equalsIgnoreCase(System.getProperty("CHECK_CONSISTENCY_ACROSS_RUNS", "false"))) {
					PrintWriter writer = null;
					
					try {
						writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream("results.log") ,"utf-8"));

						for (Map.Entry<String, Long> e: firstTotalMatchingRecordMap.entrySet()) {
							if (e.getKey().indexOf(arg0 + ":") == -1) {
								throw new RuntimeException("should never happen");
							}
							
							String queryId = e.getKey().substring(e.getKey().indexOf(':') + 1);
							writer.print("DATASET_ID=");
							writer.print(arg0);
							writer.print(":REQUEST_ID=");
							writer.print(queryId);
							writer.print(":QUEUE=");
							writer.print(dataRequestMap.get(queryId));
						
							
							if (e.getValue() != null) {
								writer.println(":TOTAL_MATCHING_RECORDS=" + e.getValue());
							} else {
								writer.println(":TOTAL_MATCHING_RECORDS=NONE");
							}
						} 
						
						writer.flush();
					} catch(IOException ioe) {
						logger.error("PERF:IO:OCCURED WHEN DUMPING RESULTS");
					} catch(Exception e) {
						logger.error("PERF:SEVERE:OCCURED WHEN DUMPING RESULTS", e);
					} finally {
						if (writer != null) {
							writer.close();
							
							if (writer.checkError()) {
								logger.error("PERF:IOE:OCCURED WHEN DUMPING RESULTS");
							}
						}
					}
				}
				synchronized(listener) {
					listener.notify();
				}
			}
		}
	}

	public void exception(String arg0, String arg1, Exception arg2) {
		if (logger.isDebugEnabled()) {
			logger.debug("ERROR_CALLBACK_EXCEPTION:" + arg0 + ":" + arg1 + arg2.getMessage(), arg2);
		} else {
			logger.error("ERROR_CALLBACK_EXCEPTION:" + arg0 + ":" + arg1 + arg2.getMessage());
		}
		
		synchronized(this) {
			exceptions++;
		}
	}

	public void httpResponseCode(String arg0, String arg1, int arg2) {
		if (arg2 != 200) {
			logger.error("ERROR_HTTP_CODE:" + arg0 + ":" + arg1 + ":HTTP_CODE=" + arg2);
			
			synchronized(this) {
				errors++;
			}
		}
	}

	public boolean isDone() {
		return count == total;
	}
	
	public String toString() {
		return "NAME=" + name + " [COUNT=" + count + ":TOTAL=" + total 
				+ ":HTTP_CODE_NOT_OK=" + errors + ":INCONSISTENT=" + this.inconsistent + ":CALLBACK_EXCEPTIONS=" + exceptions + "]";
	}
	
	public void init() {
		count = 0;
		errors = 0;
		inconsistent = 0;
		exceptions = 0;
		run++;
		totalMatchingRecordMap = Collections.synchronizedMap(new HashMap<String, Long>());
	}

	public int getRun() {
		return run;
	}
	
	public int getCount() {
		return count;
	}

	public int getErrors() {
		return errors;
	}


	public int getInconsistent() {
		return inconsistent;
	}

	public int getExceptions() {
		return exceptions;
	}
}
